/*
**	xdiff -- graphical diff for X
**	gui.c: X user-interface implementation
**
**	Copyright (C) 1996  Paul J. Lucas
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 1, or (at your option)
**	any later version.
** 
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
** 
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* system */
#include <stdio.h>
#include <stdlib.h>				/* for exit() */

/* X/Motif */
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/ScrollBar.h>

/* local */
#include "data.h"
#include "util.h"

#define	ABS(x)		( (x) < 0 ? -(x) : (x) )
#define	APP_CLASS	"XDiff"
#define	DEFAULT_FONT	"fixed"

typedef struct {				/* store all the X junk */
	Display		*display;
	char const	*geometry;
	char const	*font_name;
	XFontStruct	*font;
	GC		gc;
	struct {
		Pixel	fg, bg;
	}		color[4];		/* [ line_type ] */
	Widget		top;
	Widget		draw[2];		/* 0 = left, 1 = right */
	Widget		scroll_bar;
} x_struct;

typedef struct {				/* bundle globals */
	int		top_line;
	Dimension	window_width, window_height;
	int		line_height;
} g_struct;

/* local functions */
static void	draw PJL_PROTO(( int file_id,
			int expose_y, int expose_height
		));
static void	expose_callback PJL_PROTO(( Widget, XtPointer,
			XmDrawingAreaCallbackStruct const*
		));
static void	quit PJL_PROTO(( Widget, XEvent*, String*, Cardinal* ));
static void	resize_callback PJL_PROTO(( Widget, XtPointer,
			XmDrawingAreaCallbackStruct const*
		));
static void	scroll PJL_PROTO(( Widget, int top_line ));
static void	scrollbar_callback PJL_PROTO(( Widget, XtPointer,
			XmScrollBarCallbackStruct const*
		));

/* global variable definitions */
XtAppContext		app;

/* local variables */
static x_struct		x;
static g_struct		g;

static XtResource x_resources[] = {

	/* geometry */
	{ XtNgeometry, XtCGeometry, XtRString, sizeof( char* ),
		XtOffset( x_struct*, geometry ), XtRString, "80x40" },

	/* font */
	{ XtNfont, XtCFont, XtRString, sizeof( char* ),
		XtOffset( x_struct*, font_name ), XtRString, DEFAULT_FONT },

	/* colors for lines that are the same */
	{ "sameForeground", "SameForeground", XtRPixel, sizeof( Pixel ),
		XtOffset( x_struct*, color[ same_line ].fg ), XtRString,
		"white" },
	{ "sameBackground", "SameBackground", XtRPixel, sizeof( Pixel ),
		XtOffset( x_struct*, color[ same_line ].bg ), XtRString,
		"black" },

	/* colors for lines that are different */
	{ "diffForeground", "DiffForeground", XtRPixel, sizeof( Pixel ),
		XtOffset( x_struct*, color[ diff_line ].fg ), XtRString,
		"red" },
	{ "diffBackground", "DiffBackground", XtRPixel, sizeof( Pixel ),
		XtOffset( x_struct*, color[ diff_line ].bg ), XtRString,
		"black" },

	/* colors for lines that are different */
	{ "newForeground", "NewForeground", XtRPixel, sizeof( Pixel ),
		XtOffset( x_struct*, color[ new_line ].fg ), XtRString,
		"green" },
	{ "newBackground", "NewBackground", XtRPixel, sizeof( Pixel ),
		XtOffset( x_struct*, color[ new_line ].bg ), XtRString,
		"black" },

	/* color for "holes," lines that do not exist in the file */
	{ "holeColor", "HoleColor", XtRPixel, sizeof( Pixel ),
		XtOffset( x_struct*, color[ no_line ].fg ), XtRString,
		"gray30" }
};

static XtActionsRec x_actions[] = {
	{ "quit",	quit }
};

static String x_fallback_resources[] = {
#ifdef	NEW
	"*translations: #override\n\
		None<Key>q: quit()\n\
		Ctrl<Key>c: quit()",
#else
	"*translations: #override None<Key>q: quit()",
#endif
	NULL
};

/*************************/
	static void
draw PJL_ARG_LIST(( file_id, expose_y, expose_height ))
	PJL_ARG_DEF( int file_id )
	PJL_ARG_DEF( int expose_y )
	PJL_ARG_END( int expose_height )
{
	Window const window = XtWindow( x.draw[ file_id ] );
	register int begin, end, i, y;

	if ( expose_y + expose_height > g.window_height )
		expose_height = g.window_height - expose_y;

	begin = g.top_line + expose_y / g.line_height;
	end   = g.top_line + (expose_y + expose_height) / g.line_height;
	if ( end > map.max_lines )
		end = map.max_lines;

#ifdef	DEBUG
	printf( "draw: top_line=%d\n", g.top_line );
	printf( "draw: expose_y=%d expose_height=%d line_height=%d\n",
		expose_y, expose_height, g.line_height
	);
	printf( "draw: begin=%d end=%d\n", begin+1, end+1 );
#endif

	for (	i = begin, y = ( begin - g.top_line ) * g.line_height;
		i < end; ++i, y += g.line_height
	) {
		line_info const *line = map.set[i].line[ file_id ];
		if ( line ) {
			static char buf[ 512 + 1 ];
			int len = expand_tabs( line->text, buf, sizeof( buf ) );

			XSetForeground( x.display, x.gc,
				x.color[ line->type ].bg
			);
			XFillRectangle( x.display, window, x.gc,
				0, y, g.window_width, g.line_height
			);
			XSetForeground( x.display, x.gc,
				x.color[ line->type ].fg
			);
			XDrawString( x.display, window, x.gc,
				0, y + x.font->ascent, buf, len
			);
		} else {
			/*
			** A line in the other file does not map to a line in
			** this file so we draw a "hole."
			*/
			XSetForeground( x.display, x.gc,
				x.color[ no_line ].fg
			);
			XFillRectangle( x.display, window, x.gc,
				0, y, g.window_width, g.line_height
			);
		}
	}
}

/*************************/
	static void
expose_callback PJL_ARG_LIST(( widget, p, cbs ))
	PJL_ARG_DEF( Widget widget )
	PJL_ARG_DEF( XtPointer p )
	PJL_ARG_END( XmDrawingAreaCallbackStruct const *cbs )
{
	Window const window = XtWindow( widget );
	XExposeEvent const *const expose = &cbs->event->xexpose;
	int file_id;

	if ( expose->type == NoExpose || expose->count )
		return;

	file_id = (int)p;

	{ /* local scope */
		/*
		** Set the window background to the same color used for lines
		** that are the same to avoid an annoying flicker sometimes
		** created during scrolling using XCopyArea() when it can't
		** copy a region due to it being obscured.  When that happens,
		** it paints the region the color used by the window manager
		** which will very shortly be redrawn by us.  Between its paint
		** and our redraw is when the flicker will be noticable.
		**
		** The "proper" way to do this is by setting the background of
		** the window as soon as it's created via a CreateNotify event
		** with a SubstructureNotifyMask, but that means we'd have to
		** install an event-handler, handle the event, then uninstall
		** it.  This is too much of a pain, so we just set the
		** background color the first time this routine is called.
		*/
		static int /* bool */ set_window_bg[2];
		if ( !set_window_bg[ file_id ] ) {
			XSetWindowBackground( x.display, window,
				x.color[ same_line ].bg
			);
			XClearWindow( x.display, window );
			set_window_bg[ file_id ] = 1 /* true */;
		}
	} /* local scope */

	draw( (int)p, expose->y, expose->height );
}

/*************************/
	void
gui_output()
{
	Widget form;
	XGCValues gc_v;
	unsigned width, height;
	int i;

	/*
	** Create the top-level shell, load the application resources and font.
	*/
	x.top = XtAppCreateShell( me, APP_CLASS,
		applicationShellWidgetClass, x.display, NULL, 0
	);
	XtGetApplicationResources( x.top, &x,
		x_resources, XtNumber( x_resources ), NULL, 0
	);
	if (	!( x.font = XLoadQueryFont( x.display, x.font_name ) ) &&
		!( x.font = XLoadQueryFont( x.display, DEFAULT_FONT ) )
	) {
		fprintf( stderr, "%s: can't load font \"%s\"\n",
			me, DEFAULT_FONT
		);
		exit( 1 );
	}
	g.line_height = x.font->ascent + x.font->descent;

	/*
	** We want to parse the geometry specification ourselves to interpret
	** <width>x<height> as <colums>x<rows> like an xterm does.  However, we
	** want <colums> to mean the number of columns for each file.  So, for
	** a geometry specification of 80x24, the window would be 160x24.
	**
	** We needn't check the return value of XParseGeometry() since the
	** width and height will always have been specified falling back on our
	** application resources if the geometry is not specified elsewhere.
	*/
	XParseGeometry( x.geometry, &i, &i, &width, &height );

	/*
	** Below, we set XmNheightInc to allow changes in the window's height
	** only in multiples of line_height.  Once done, the <height> in the
	** geometry specification is automatically correctly interpreted as
	** <rows>.  We don't set XmNwidthInc, even though we'd like to, since
	** the presence of the scroll bar complicates the calculation.  Because
	** we don't set it, the auto-correction of the geometry width doesn't
	** happen.  As a result, we convert <columns> to width-in-pixels
	** ourselves using the calculation below.  The additional 2% is for the
	** scroll bar.
	**
	** No, I don't know why XmNheight and XmNmaxHeight have "+ 1" in them,
	** but it makes it work properly.
	*/
	width *= 2 * x.font->max_bounds.width * 1.02;

	XtVaSetValues( x.top,
		XmNbaseHeight,	1,
		XmNheight,	g.line_height * (height + 1),
		XmNheightInc,	g.line_height,
		XmNminHeight,	g.line_height * 10,
		XmNmaxHeight,	g.line_height * (map.max_lines + 1),

		XmNbaseWidth,	1,
		XmNwidth,	width,
		XmNminWidth,	500,
		/*
		** Since we handle the geometry specification, we have to zap
		** the original one by setting XmNgeometry to NULL otherwise it
		** will take precedence and ignore all of our calculations.
		*/
		XmNgeometry,	NULL,
		NULL
	);

	/*
	** Create the widgets.
	*/
	form = XtVaCreateManagedWidget( "form",
		xmFormWidgetClass,		x.top,
		XmNfractionBase,		100,
		NULL
	);

	x.scroll_bar = XtVaCreateManagedWidget( "scrollBar",
		xmScrollBarWidgetClass,		form,
		XmNtopAttachment,		XmATTACH_FORM,
		XmNbottomAttachment,		XmATTACH_FORM,
		XmNleftAttachment,		XmATTACH_POSITION,
		XmNleftPosition,		50,
		XmNleftOffset,			-8,
		XmNminimum,			0,
		XmNmaximum,			map.max_lines,
		XmNincrement,			1,
		NULL
	);
	XtAddCallback( x.scroll_bar, XmNvalueChangedCallback,
		(XtCallbackProc)&scrollbar_callback, NULL
	);
	XtAddCallback( x.scroll_bar, XmNdragCallback,
		(XtCallbackProc)&scrollbar_callback, NULL
	);

	x.draw[0] = XtVaCreateManagedWidget( "leftDraw",
		xmDrawingAreaWidgetClass,	form,
		XmNtopAttachment,		XmATTACH_FORM,
		XmNbottomAttachment,		XmATTACH_FORM,
		XmNleftAttachment,		XmATTACH_FORM,
		XmNrightAttachment,		XmATTACH_WIDGET,
		XmNrightWidget,			x.scroll_bar,
		NULL
	);
	/*
	** Add the resize callback to only one of the two drawing areas so when
	** the window is resized, we don't call the callback twice.
	*/
	XtAddCallback( x.draw[0], XmNresizeCallback,
		(XtCallbackProc)&resize_callback, (XtPointer)0
	);
	x.draw[1] = XtVaCreateManagedWidget( "rightDraw",
		xmDrawingAreaWidgetClass,	form,
		XmNtopAttachment,		XmATTACH_FORM,
		XmNbottomAttachment,		XmATTACH_FORM,
		XmNleftAttachment,		XmATTACH_WIDGET,
		XmNleftWidget,			x.scroll_bar,
		XmNrightAttachment,		XmATTACH_FORM,
		NULL
	);

	for ( i = 0; i <= 1; ++i ) {
		/*
		** Make it so that both multiple Expose and GraphicsExpose
		** events in the event queue are compressed into a single
		** event.
		*/
		x.draw[i]->core.widget_class->core_class.compress_exposure =
			XtExposeCompressMultiple |
			XtExposeGraphicsExposeMerged;

		XtAddCallback( x.draw[i], XmNexposeCallback,
			(XtCallbackProc)&expose_callback, (XtPointer)i
		);
	}

	/*
	** Set the font in the GC.
	*/
	gc_v.font = x.font->fid;
	x.gc = XCreateGC( x.display, RootWindowOfScreen( XtScreen( x.top ) ),
		GCFont, &gc_v
	);

	/*
	** Other initializations.
	*/
	g.top_line = 0;
	XtVaGetValues( x.draw[0],
		XmNwidth,	&g.window_width,
		XmNheight,	&g.window_height,
		NULL
	);

	XtRealizeWidget( x.top );
	XtAppMainLoop( app );
}

/*************************/
	void
init_gui PJL_ARG_LIST(( p_argc, argv ))
	PJL_ARG_DEF( int *p_argc )
	PJL_ARG_END( char const *argv[] )
{
	/*
	** We have to initialize the X toolkit the hard way, i.e., not use
	** XtAppInitialize(), since we have to insert the call to
	** XtAppAddActions() at just the right spot.
	*/
	XtToolkitInitialize();
	app = XtCreateApplicationContext();
	XtAppAddActions( app, x_actions, XtNumber( x_actions ) );
	XtAppSetFallbackResources( app, x_fallback_resources );
	x.display = XtOpenDisplay( app, NULL, me, APP_CLASS, NULL, 0,
		p_argc, (char**)argv
	);
}

/*************************/
	static void
quit PJL_ARG_LIST(( widget, event, params, num_params ))
	PJL_ARG_DEF( Widget widget )		/* unused */
	PJL_ARG_DEF( XEvent *event )		/* unused */
	PJL_ARG_DEF( String *params )		/* unused */
	PJL_ARG_END( Cardinal *num_params )	/* unused */
{
	exit( 0 );
}

/*************************/
	static void
resize_callback PJL_ARG_LIST(( widget, p, cbs ))
	PJL_ARG_DEF( Widget widget )
	PJL_ARG_DEF( XtPointer p )		/* unused */
	PJL_ARG_END( XmDrawingAreaCallbackStruct const *cbs )
{
	/*
	** Get the new window_width and window_height and adjust the scroll bar
	** appropriately.  Expose event(s) will be generated automatically if
	** needed.
	*/
	int lines_in_window;

	XtVaGetValues( widget,
		XmNwidth,	&g.window_width,
		XmNheight,	&g.window_height,
		NULL
	);

	lines_in_window = g.window_height / g.line_height;
#ifdef	DEBUG
	printf( "resize: lines_in_window=%d\n", lines_in_window );
#endif

	/*
	** If the bottom of the window was stretched such that there are
	** insufficient lines to draw in the newly-exposed region, we have to
	** scroll the window down (text up) to compensate.
	*/
	if ( g.top_line && map.max_lines - g.top_line < lines_in_window ) {
#ifdef	DEBUG
		printf( "resize: max_lines=%d top_line=%d\n",
			map.max_lines, g.top_line
		);
#endif
		scroll( widget, map.max_lines - lines_in_window );
	}

	/*
	** Don't let the slider values exceed the maximum.
	*/
	if ( lines_in_window > map.max_lines )
		lines_in_window = map.max_lines;

	XtVaSetValues( x.scroll_bar,
		XmNvalue,		g.top_line,
		XmNsliderSize,		lines_in_window,
		XmNpageIncrement,	lines_in_window - 1,
		NULL
	);
}

/*************************/
	static void
scroll PJL_ARG_LIST(( widget, top_line ))
	PJL_ARG_DEF( Widget widget )
	PJL_ARG_END( int top_line )
{
	int delta = ( top_line - g.top_line ) * g.line_height;
	int y_clr, scroll_down;

	if ( !delta )
		return;

	g.top_line = top_line;

	/* note: by "scrolling down" we actually scroll the text up! */
	scroll_down = delta > 0;
	delta = ABS( delta );

	y_clr = 0;
	if ( delta < g.window_height ) {
		int y_src, y_dst, i;

		/*
		** We scrolled less than a full screen; therefore, it makes
		** sense to scroll (copy) the text.
		*/
		if ( scroll_down ) {
			y_src = delta;
			y_dst = 0;
			y_clr = g.window_height - delta;
		} else {
			y_src = 0;
			y_dst = delta;
		}

		for ( i = 0; i <= 1; ++i ) {
			Window const window = XtWindow( x.draw[ i ] );
			XCopyArea( x.display, window, window, x.gc,
				0, y_src,
				g.window_width, g.window_height - delta,
				0, y_dst
			);
		}
	}

	/*
	** Draw the "exposed" lines directly rather than generating an Expose
	** event -- this is faster.
	*/
	draw( 0, y_clr, delta );
	draw( 1, y_clr, delta );
}

/*************************/
	static void
scrollbar_callback PJL_ARG_LIST(( widget, p, cbs ))
	PJL_ARG_DEF( Widget widget )
	PJL_ARG_DEF( XtPointer p )		/* unused */
	PJL_ARG_END( XmScrollBarCallbackStruct const *cbs )
{
	scroll( widget, cbs->value );
}
